#include <stdio.h>

#define NULLFILE ((FILE *)NULL)

#define REDBOOK_START		0
#define REDBOOK_LENGTH		0x40000

#define CONF_SECTION_LENGTH	0x20000
#define CONF_HEADER		(REDBOOK_START + REDBOOK_LENGTH)
#define CONF_HEADER_LENGTH	0x10
#define CONF_START		(CONF_HEADER + CONF_HEADER_LENGTH)
#define CONF_LENGTH		(CONF_SECTION_LENGTH - CONF_HEADER_LENGTH)

#define KERNEL_SECTION_LENGTH	0x100000
#define KERNEL_HEADER		(CONF_START + CONF_LENGTH)
#define KERNEL_HEADER_LENGTH	0x10
#define KERNEL_START		(KERNEL_HEADER + KERNEL_HEADER_LENGTH)
#define KERNEL_LENGTH		(KERNEL_SECTION_LENGTH - KERNEL_HEADER_LENGTH)

#define TRAILER_OFFSET		(ROMFS_START + ROMFS_LENGTH)
#define TRAILER_LENGTH		0x10

#define ROMFS_SECTION_LENGTH    (0x6A0000 - TRAILER_LENGTH)
#define ROMFS_HEADER            (KERNEL_START + KERNEL_LENGTH)
#define ROMFS_HEADER_LENGTH     0x10
#define ROMFS_START             (ROMFS_HEADER + ROMFS_HEADER_LENGTH)
#define ROMFS_LENGTH            (ROMFS_SECTION_LENGTH - ROMFS_HEADER_LENGTH)

#define IMAGE_SIZE		0x800000

long findLength (FILE *fpin, long offset, long len)
{
int k;
int c;
long l = 0;

	fseek (fpin, offset, SEEK_SET);
	l += (fgetc(fpin) << 24);
	l += (fgetc(fpin) << 16);
	l += (fgetc(fpin) << 8);
	l += (fgetc(fpin));
	for (k = 4; k < len; k++)
		fgetc (fpin);

	return l;
}


long writefile (long len, FILE *fpin, FILE *fpout)
{
long k;
int c;

	for (k = 0; k < len; k++)	{
		c = fgetc (fpin);
		if (c == EOF)
			break;
		fputc (c, fpout);
	}
	return k;
}


void usage (void)
{
	printf ("Usage: slugtool [pack|unpack|size] <imagefile>\n");
	exit (-1);
}

#define ALTER_HEADER_LENGTH	11732
#define ALTER_TRAILER_LENGTH	112

void size (char *which)
{
FILE *fpin;
long l;

	fpin = fopen ("vmlinuz", "r");
	fseek (fpin, 0, SEEK_END);
	l = ftell (fpin);
	l -= ALTER_HEADER_LENGTH;
	l -= ALTER_TRAILER_LENGTH;
	fclose (fpin);
	if (!strcmp (which, "kernel"))
		printf ("%ld\n", l);
	else if (!strcmp (which, "main"))	{
		l += ALTER_HEADER_LENGTH;
		printf ("%ld\n", l);
	}
}

void unpack (char *filename)
{
FILE *fpin, *fpout;
long l;
	printf ("Unpacking '%s'....\n", filename);
	fpin = fopen (filename, "r");

	fseek (fpin, REDBOOK_START, SEEK_SET);
	printf ("\tRedbook....\n");
	fpout = fopen ("Redbook", "w");
	writefile (REDBOOK_LENGTH, fpin, fpout);
	fclose (fpout);

	printf ("\tReading SysConf Header....\n");
	l = findLength (fpin, CONF_HEADER, CONF_HEADER_LENGTH);

	fseek (fpin, CONF_START, SEEK_SET);
	printf ("\tSysConf....\n");
	fpout = fopen ("SysConf", "w");
	writefile (l, fpin, fpout);
	fclose (fpout);

	printf ("\tReading vmlinuz Header....\n");
	l = findLength (fpin, KERNEL_HEADER, KERNEL_HEADER_LENGTH);

	fseek (fpin, KERNEL_START, SEEK_SET);
	printf ("\tvmlinuz....\n");
	fpout = fopen ("vmlinuz", "w");
	writefile (l, fpin, fpout);
	fclose (fpout);

	printf ("\tReading ramdisk Header....\n");
	l = findLength (fpin, ROMFS_HEADER, ROMFS_HEADER_LENGTH);

	fseek (fpin, ROMFS_START, SEEK_SET);
	printf ("\tramdisk.gz....\n");
	fpout = fopen ("ramdisk.gz", "w");
	writefile (l, fpin, fpout);
	fclose (fpout);

	fseek (fpin, TRAILER_OFFSET, SEEK_SET);
	printf ("\tTrailer....\n");
	fpout = fopen ("Trailer", "w");
	writefile (TRAILER_LENGTH, fpin, fpout);
	fclose (fpout);

	fclose (fpin);
}

void blankHeader (FILE *fpout, int length)
{
int k;

	for (k = 0; k < length; k++)
		fputc (0, fpout);
}

void storeLength (FILE *fpout, long offset, long len)
{
	fseek (fpout, offset, SEEK_SET);
	fputc ( (len >> 24) & 0x00ff, fpout);
	fputc ( (len >> 16) & 0x00ff, fpout);
	fputc ( (len >> 8) & 0x00ff, fpout);
	fputc ( len  & 0x00ff, fpout);
}


void pack (char *filename)
{
FILE *fpin, *fpout;
char buffer[1024];


	printf ("Packing '%s'....\n", filename);
	printf ("\tzeroing entire file....\n");
	sprintf (buffer, "dd if=/dev/zero of=%s ibs=%ld count=1 2>/dev/null", filename, IMAGE_SIZE);
	system (buffer);

	fpout = fopen (filename, "r+");

	printf ("\tRedbook....\n");
	fpin = fopen ("Redbook", "r");
	if (fpin == NULLFILE) {
		printf ("File 'Redbook' not found!\n");
		usage ();
	}
	fseek (fpout, REDBOOK_START, SEEK_SET);
	writefile (REDBOOK_LENGTH, fpin, fpout);
	fclose (fpin);

	printf ("\tSysConf....\n");
	fpin = fopen ("SysConf", "r");
	if (fpin == NULLFILE) {
		printf ("File 'SysConf' not found!\n");
		usage ();
	}
	fseek (fpout, CONF_HEADER, SEEK_SET);
	blankHeader (fpout, CONF_HEADER_LENGTH);
	fseek (fpout, CONF_START, SEEK_SET);
	storeLength (fpout, CONF_HEADER, writefile (CONF_LENGTH, fpin, fpout));
	fclose (fpin);

	printf ("\tvmlinuz....\n");
	fpin = fopen ("vmlinuz", "r");
	if (fpin == NULLFILE) {
		printf ("File 'vmlinuz' not found!\n");
		usage ();
	}
	fseek (fpout, KERNEL_HEADER, SEEK_SET);
	blankHeader (fpout, KERNEL_HEADER_LENGTH);
	fseek (fpout, KERNEL_START, SEEK_SET);	
	storeLength (fpout, KERNEL_HEADER, writefile (KERNEL_LENGTH, fpin, fpout));
	fclose (fpin);

	printf ("\tramdisk.gz....\n");
	fpin = fopen ("ramdisk.gz", "r");
	if (fpin == NULLFILE) {
		printf ("File 'ramdisk.gz' not found!\n");
		usage ();
	}
	fseek (fpout, ROMFS_HEADER, SEEK_SET);
	blankHeader (fpout, ROMFS_HEADER_LENGTH);
	fseek (fpout, ROMFS_START, SEEK_SET);	
	storeLength (fpout, ROMFS_HEADER, writefile (ROMFS_LENGTH, fpin, fpout));
	fclose (fpin);

	printf ("\tTrailer....\n");
	fpin = fopen ("Trailer", "r");
	if (fpin == NULLFILE) {
		printf ("File 'Trailer' not found!\n");
		usage ();
	}
	fseek (fpout, TRAILER_OFFSET, SEEK_SET);
	writefile (TRAILER_LENGTH, fpin, fpout);
	fclose (fpin);

	fclose (fpout);
}

int
main (int argc, char *argv[])
{
	if (argc < 3)
		usage ();

	if (!strcmp (argv[1], "unpack"))
		unpack (argv[2]);
	else if (!strcmp (argv[1], "pack"))
		pack (argv[2]);
	else if (!strcmp (argv[1], "size"))
		size (argv[2]);
	else
		usage ();
	return 0;
}
